/*
** Module   : SymbolResolver
** Abstract : Enables compile-time symbol resolution for expressions
**
** Copyright (c) 2001-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 ECF 20041130   @19000 Production version from JTA, reworked
**                           into a new package and with minor cleanups
**                           in javadoc.
** 002 ECF 20041210   @19001 Fixed typos and code cleanup.
** 003 ECF 20050201   @19678 Overhauled from predecessor VariableResolver
**                           interface. Added new methods for first class
**                           function callback support and first class
**                           Boolean support.
** 004 ECF 20050215   @19826 Replaced querySymbolType(String, Object[])
**                           with queryReturnType(String, Object[]). The
**                           former returned an integer constant; the
**                           latter returns a Class object indicating the
**                           return type. This simplifies processing
**                           functions which contain nested functions as
**                           parameters.
** 005 ECF 20050310   @20269 Added resolveToLiteral method. Needed to
**                           support parser-level resolution of string
**                           constants to numeric, boolean, or string
**                           literals.
** 006 ECF 20050422   @20849 Changed resolveToLiteral method signature. It
**                           now accepts an optional namespace parameter.
** 007 ECF 20050505   @21355 Complete rewrite. Overhauled from an
**                           interface to the abstract base class of all
**                           expression symbol resolver concrete
**                           implementations. Now supports object
**                           references directly, manages a variable pool
**                           transparently, and adds new security
**                           features. All variable and library resolution
**                           is performed at parse/compile time, and no
**                           symbol resolution calls are made to the
**                           resolver at expression execution time.
** 008 ECF 20050608   @21455 Fixed NPE in CacheKey constructor. Signature
**                           hash calculation was not checking for null
**                           parameter types.
** 009 ECF 20060219   @24706 Minor optimization. Iterate variable pool map
**                           values instead of keys.
** 010 GES 20070628   @34318 Moved the compiled expression instance cache
**                           here from the Expression class where it was
**                           a static. This class' purpose is to store
**                           data across expressions so it is a more
**                           appropriate location. In addition, since the
**                           cache can now be instance data, this can
**                           greatly reduce the working set of a process
**                           which loads more than 1 expression "engine"
**                           sequentially. The previous approach pinned
**                           all expressions, the symbol resolver, all
**                           libraries... in memory forever! This cache
**                           is exposed via new get/put methods and
**                           internally, the access to the cache is
**                           protected via granular synchronization which
**                           should make this thread safe.
** 011 ECF 20070629   @34350 Minor enhancements. Made CacheKey members
**                           final. Replaced StringBuffer with
**                           StringBuilder.
** 012 ECF 20080730   @39261 Changed expression caching and management of
**                           libraries and variables. Compiled expressions
**                           are now tiered in the cache by scope, so that
**                           all instances associated with a given scope
**                           can be released when that scope ends.
**                           Individual variables and function libraries
**                           can be dropped.
** 013 GES 20130203          Modified variable registration to utilize a new
**                           Initializer class. This allows a hard coded
**                           instance to be set into a variable in addition
**                           to the current expression evaluation mechanism.
*/

package com.goldencode.expr;

import java.lang.ref.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.concurrent.locks.*;

/**
 * The abstract base class of application-specific symbol resolvers. This
 * class provides the link between the expression engine and client code.
 * It is responsible for resolving symbols found in user expressions into
 * objects and user functions. In addition, it is responsible for enforcing
 * security with respect to the accessibility of particular classes and
 * methods from within user expressions. This class employs a callback model
 * to accomplish these tasks.
 * <p>
 * <b>General Use</b><br>
 * In order to enable full support for user expressions which are sensitive
 * to the context of the application, client code must register one or more
 * objects as callback libraries before expressions are submitted to the
 * engine for processing. This is done using the {@link #registerLibrary}
 * method, which exposes the public methods of the registered object's class
 * to user expressions. Once a library object is registered, the following
 * user expression library constructs/concepts are managed at this abstract
 * implementation level:
 * <ul>
 * <li><b>Property Access.</b> If a registered library's class contains any
 *     bean-like method signatures (<code>get/isXXXX()</code> or
 *     <code>setXXXX(&lt;property&gt;)</code>), these methods are used to
 *     expose the backing property to user expressions. Each property can
 *     then be used in an expression by its simple name for access and
 *     assignment purposes (e.g., <code>"Larry" == name</code> or
 *     <code>stooge.name = "Moe"</code>). Properties may be referenced in
 *     an expression either with a library qualifier or without. In the
 *     latter case, the property name must be unambiguous among all
 *     registered libraries, else an exception will be thrown at expression
 *     compile time.
 * <li><b>User Function Calls.</b> Public methods of a registered library
 *     which do not have bean-like signatures are exposed to user expressions
 *     as user functions. These are accessed using standard method invocation
 *     syntax, either qualified by a symbol representing the implementing
 *     library, or unqualified. If unqualified, the user function name and
 *     signature must be unambiguous among all registered libraries, else an
 *     exception will be thrown at expression compile time.
 * <li><b>Method Invocation.</b> Methods may be invoked on constants and
 *     variables using the familiar syntax:
 *     <code>&lt;target&gt;.&lt;method&gt;([argument-list...])</code>.
 *     It is not necessary for the concrete symbol resolver implementation
 *     to know how to resolve or invoke these types of method calls;
 *     however, it will have an opportunity to deny permission for an
 *     expression to access any class or invoke any method via the built-in
 *     security mechanism.
 * </ul>
 * <p>
 * In addition to callback libraries, the symbol resolver manages variable
 * pools, segregated by <a href="#scope">scope</a> and within scope, by name.
 * A variable has a data type, which is declared at the time it is {@link
 * #registerVariable registered}. Optionally, it also has an initializer
 * expression associated with it, which is used to set the value of the
 * variable whenever it is {@link #resetVariables reset}.
 * <p>
 * A {@link Variable} is referenced in user expressions by its name. During
 * expression parsing, a variable's name symbol is resolved into one of the
 * following method invocations:
 * <ul>
 * <li>{@link CompiledExpression#getVar} - used if the value of the variable
 *     is being retrieved in the context of the expression (e.g.,
 *     <code>myVar + 50 == 100</code>).
 * <li>{@link CompiledExpression#setVar} - used if the value of the variable
 *     is being set in the context of the expression, via an assignment
 *     (e.g., <code>myVar = 100 - 50</code>).
 * </ul>
 * <p>
 * Variable pools are maintained transparently by the expression engine.
 * References to variable names in expressions are resolved using the
 * resolver's {@link #getCurrentScope current scope}, and any enclosing
 * scopes if not found in the current scope. Thus, variables must be
 * registered within a defined scope. Variables may contain simple literal
 * values or more complex objects. It is legal to invoke a method against
 * a variable symbol in a user expression. It is <em>not</em> necessary to
 * unwrap primitive literals from their Java wrappers, such as
 * <code>Long</code>, <code>Integer</code>, <code>Boolean</code>, and
 * <code>Double</code>. Autoboxing and some simple type conversion is done
 * automatically by the expression engine.
 * <p>
 * A concrete symbol resolver implementation is responsible for managing
 * only one construct in a user expression: constants. Constants are symbols
 * which are resolved at expression parse time into simple objects. These
 * objects represent numeric or string literals. Although a constant is an
 * object, it is illegal to invoke a method against a constant symbol in a
 * user expression. Constants are resolved when the expression engine calls
 * the {@link #resolveConstant} method. The default implementation of this
 * method returns <code>null</code>. Subclasses should override this method
 * if it is necessary to support an application-specific set of constant
 * values.
 * <p>
 * Care should be taken when choosing constant and variable names to avoid
 * conflicts with the names of properties exposed via registered callback
 * libraries. The order of precedence when resolving these is:
 * <ol>
 * <li>A constant hides properties and variables which use the same symbol.
 * <li>A property hides variables which use the same symbol.
 * <li>A variable hides nothing.
 * </ol>
 * <p>
 * These conflicts <em>are not</em> detected automatically by the expression
 * engine. A conflict will manifest itself as the <em>lack</em> of a call to
 * a getter/setter method (for properties) or the <em>lack</em> of a request
 * to resolve a symbol name to an object (for variables).
 * <p>
 * <a name="scope"><b>Scope</b></a><br>
 * A concrete implementation of this class is also responsible for defining
 * a scoping mechanism. Scopes are used to segregate variable pools and
 * library reference mappings, such that sibling scopes can each have
 * variables and libraries mapped to the same names without producing
 * conflicts. Scopes may be nested, which allows a user expression to refer
 * to variables and libraries defined for scopes which enclose the current
 * scope. The details of how scopes are arranged and defined are completely
 * up to the client code layer. The mechanism defined by the client is
 * shared with the expression engine through several methods:
 * <ul>
 * <li>{@link #registerLibrary}
 * <li>{@link #registerVariable}
 * <li>{@link #getCurrentScope}
 * <li>{@link #resetVariables}
 * <li>{@link Scope#getEnclosingScope}
 * </ul>
 *
 * In each case, an instance of {@link Scope} is used to define the context
 * of the libraries and/or variables being manipulated.
 * <p>
 * <b>Important:</b> it is critical that certain method invocations on {@link
 * Expression} be allowed to complete within the context of a single scope.
 * The scope should not change while any of the following method invocations
 * are in progress:
 * <ul>
 * <li>{@link Expression#execute}
 * <li>{@link Expression#getReturnType}
 * </ul>
 * This is because these methods may trigger expression parsing and
 * compilation, which requires a consistent scope be maintained while all
 * symbols encountered within the expression are properly resolved.
 * <b>Failure to observe this requirement may cause unpredictable and
 * unwanted results!</b>
 * <p>
 * <b>Security Considerations</b><br>
 * The ability to invoke arbitrary methods from within a user expression has
 * obvious security implications. Security is enforced by an implementation
 * of this class by overriding the {@link #checkPermission} method. This
 * callback is invoked by the expression engine during parsing of an
 * expression. As method invocations are detected, this hook is used to
 * request permission to allow the method to be invoked at runtime. In the
 * event that the method is detected to be within a registered library, the
 * <code>registered</code> parameter of <code>checkPermission</code> is
 * set to <code>true</code>; otherwise, it is set to <code>false</code>.
 * <b>No additional runtime check is made, so if a method invocation
 * is permitted at parse time, it will be permitted each time the
 * expression is executed.</b>.
 * <p>
 * <b>By default, security is disabled.</b> To enable the permission check
 * mechanism, subclasses <b>must</b> override the <code>checkPermission</code>
 * method.
 */
public abstract class SymbolResolver
implements CompilerConstants
{
   /** Static method within CompiledExpression to get library at runtime */
   private static final Method METH_GET_LIB;
   
   /** Static method within CompiledExpression to get variable at runtime */
   private static final Method METH_GET_VAR;
   
   /** Static method within CompiledExpression to set variable at runtime */
   private static final Method METH_SET_VAR;
   
   // Initialize above variables for CompiledExpression methods.
   static
   {
      Class<CompiledExpression> exprClass = CompiledExpression.class;
      Class<?>[] sigGet = { Integer.TYPE };
      Class<?>[] sigSet = { Integer.TYPE, Object.class };
      
      try
      {
         METH_GET_LIB = exprClass.getDeclaredMethod("getLib", sigGet);
         METH_GET_VAR = exprClass.getDeclaredMethod("getVar", sigGet);
         METH_SET_VAR = exprClass.getDeclaredMethod("setVar", sigSet);
      }
      catch (NoSuchMethodException exc)
      {
         throw new RuntimeException(exc);
      }
   }
   
   /** Compiled expression instance cache */
   private static final
   Map<Scope, Map<Object, CompiledExpression>> instCache =
      new WeakHashMap<Scope, Map<Object, CompiledExpression>>();
   
   /** Mutex lock on compiled expression instance cache */
   private static final ReentrantLock instCacheLock = new ReentrantLock();
   
   /** Map of libraries by scope, name; used at parse and compile time */
   private final Map libraries = new WeakHashMap();
   
   /** Map of user variables by scope, name; used at parse and compile time */
   private final Map variables = new WeakHashMap();
   
   /** Map of library classes by scope */
   private final Map libraryClasses = new WeakHashMap();
   
   /** Library targets by scope, then by class */
   private final Map<Scope, Map<Class<?>, Object>> libsByClass =
      new WeakHashMap<Scope, Map<Class<?>, Object>>();
   
   /** Cache of functions, used for quick method lookups during parsing */
   private SoftReference cacheRef = new SoftReference(null);
   
   /** Current scope for library and variable resolution */
   private Scope currentScope = null;
   
   /**
    * Return an implementation of this abstract class for test and debugging
    * purposes.
    *
    * @return  Instance of a test <code>SymbolResolver</code> implementation.
    */
   static SymbolResolver getTestResolver()
   {
      return new TestResolver();
   }
   
   /**
    * Register the target object as a runtime callback library. Registration
    * makes all public methods of the target object's class accessible to
    * expressions. At expression parse time, symbols are resolved to these
    * methods. Symbols may represent direct method calls using common Java
    * method call syntax. In addition, symbols which are represented in a
    * library by Java-bean style accessor methods (get/is... and set...)
    * can be included in an expression as the name of the property directly.
    * So, for instance, the expression <code>lib.format = "zz9"</code> is
    * an acceptable notation which would translate to a library call of
    * <code>lib.setFormat("zz9")</code>, and <code>test(lib.format)</code>
    * would translate to <code>test(lib.getFormat())</code>, where
    * <code>public void setFormat(String)</code> and <code>public String
    * getFormat()</code> are members of the registered library named
    * <code>lib</code>.
    *
    * @param   scope
    *          Scope associated with the library being registered. May not
    *          be <code>null</code>.
    * @param   name
    *          Name associated with the library being registered. If
    *          provided, this name may be used in expressions to qualify
    *          references to methods and properties (exposed via Java-bean
    *          style getters/setters) in the registered library. Only one
    *          library target may be registered to this name per scope.
    *          A single library per scope can be registered with a
    *          <code>null</code> name.
    * @param   target
    *          Target object whose class implements public methods which
    *          are exposed to expressions.
    *
    * @throws  NullPointerException
    *          if <code>scope</code> is <code>null</code>.
    * @throws  AmbiguousSymbolException
    *          if <code>name</code> is already in use within the specified
    *          scope, or if an instance of <code>target</code>'s class is
    *          already registered in this scope.
    *
    * @see     #getCurrentScope
    */
   public final synchronized void registerLibrary(Scope scope,
                                                  String name,
                                                  Object target)
   {
      // Scope must not be null.
      if (scope == null)
      {
         throw new NullPointerException("Scope must not be null");
      }
      
      // Map target object by name within scope.
      Map library = (Map) libraries.get(scope);
      if (library == null)
      {
         library = new HashMap();
         libraries.put(scope, library);
      }
      
      // Name must not be in use for this scope.
      if (library.containsKey(name))
      {
         throw new AmbiguousSymbolException(
            "Cannot register duplicate library '" + name +
            "' in scope:  " + scope);
      }
      
      // Target's class must not already be registered in this scope.
      Map map = (Map) libraryClasses.get(scope);
      if (map == null)
      {
         map = new HashMap();
         libraryClasses.put(scope, map);
      }
      Class targetClass = target.getClass();
      if (map.containsKey(targetClass))
      {
         throw new AmbiguousSymbolException(
            "Cannot register duplicate library type (" +
            targetClass.getName() + ") in scope:  " + scope);
      }
      
      // Register target in this scope.
      library.put(name, target);
      
      // Add target's class to map within scope.
      map.put(targetClass, name);
      
      // Register target by class within scope.
      Map<Class<?>, Object> library2 = libsByClass.get(scope);
      if (library2 == null)
      {
         library2 = new HashMap<Class<?>, Object>(4);
         libsByClass.put(scope, library2);
      }
      library2.put(targetClass, target);
   }
   
   /**
    * Drop a particular function library from the scope of libraries
    * registered at the given scope.  This permanently removes the library.
    * 
    * @param   scope
    *          Scope with which target library is associated.
    * @param   name
    *          Name of library to be dropped.
    */
   public final synchronized void dropLibrary(Scope scope, String name)
   {
      Map library = (Map) libraries.get(scope);
      if (library == null)
      {
         return;
      }
      
      Object target = library.remove(name);
      if (target == null)
      {
         return;
      }
      
      Class<?> targetClass = target.getClass();
      
      Map map = (Map) libraryClasses.get(scope);
      map.remove(targetClass);
      
      Map library2 = libsByClass.get(scope);
      library2.remove(targetClass);
   }
   
   /**
    * Create a new user variable and register it for use within expressions.
    * A variable must have a name and a scope. User expressions may refer
    * to variables by name to either retrieve the value of the variable
    * (e.g., <code>myVar + 50 == 100</code>), or to set the value of the
    * variable using assignment syntax (e.g., <code>myVar = 100 - 50</code>).
    * If a variable is registered as read-only, the latter syntax is not
    * permitted, however. Variables may also have methods invoked upon them.
    * The methods permitted are those supported by the variable's data type.
    * If a variable represents an object wrapper which contains a primitive
    * value (e.g., an instance of <code>java.lang.Integer</code>), it is
    * <em>not</em> necessary to invoke any method to unwrap the primitive
    * (e.g. <code>myVar.intValue()</code>) value in order to use it.
    * Autoboxing and some simple type conversions are performed automatically
    * by the expression engine when compiling the expression.
    *
    * @param   scope
    *          Scope associated with the variable being registered. May not
    *          be <code>null</code>.
    * @param   name
    *          The name under which the variable is registered in this scope.
    *          Must not be <code>null</code>.
    * @param   type
    *          The data type of the variable. If <code>null</code>, the
    *          return type of <code>initializer</code> is used. If not
    *          <code>null</code>, it must be assignment compatible
    *          (considering autoboxing and simple type conversions) with
    *          the return type of <code>initializer</code>. If
    *          <code>initializer</code> is <code>null</code>, this parameter
    *          is required.
    * @param   initializer
    *          A helper which is used to initialize the value of the
    *          new variable. See description of <code>type</code> parameter
    *          for details on how this helper's return type is considered
    *          with respect to the variable's data type. This parameter may
    *          be <code>null</code> if <code>type</code> is not. If no
    *          initializer is provided, the variable's value will be
    *          initialized to <code>null</code> whenever the variable is
    *          {@link Variable#reset reset}. The initializer is applied
    *          immediately during construction of the variable, so any
    *          libraries or variables it references must already be
    *          registered. It should be assumed that any initializer
    *          expression for a variable will execute in the same scope as
    *          is defined for the variable.
    * @param   readOnly
    *          If <code>true</code>, this variable's value may not be set
    *          from within an expression. If <code>false</code>, assignments
    *          to this variable in an expression are permitted. Regardless
    *          of whether a variable is registered as read-only, its value
    *          may always be changed programmatically using {@link
    *          Variable#setValue}.
    *
    * @throws  NullPointerException
    *          if either <code>scope</code> or <code>name</code> is
    *          <code>null</code>.
    * @throws  SymbolException
    *          if <code>name</code> is already registered to a variable
    *          declared in <code>scope</code>; if <code>type</code>
    *          represents a primitive class; if <code>type</code> was
    *          <code>null</code> and could not be determined from
    *          <code>initializer</code>'s return type; if <code>type</code>
    *          was not <code>null</code> and was incompatible with
    *          <code>initializer</code>'s return type.
    *
    * @return  The newly registered variable. The caller should store this
    *          if direct access to the <code>Variable</code> object will be
    *          required after registration (e.g., to invoke the object's
    *          {@link Variable#setValue} method directly). An alternate
    *          method to retrieve the variable object after registration is
    *          the {@link #getVariable} method.
    */
   public final synchronized Variable registerVariable(Scope scope,
                                                       String name,
                                                       Class<?> type,
                                                       Initializer initializer,
                                                       boolean readOnly)
   {
      // Scope must not be null.
      if (scope == null)
      {
         throw new NullPointerException("Scope must not be null");
      }
      
      // Name must not be null.
      if (name == null)
      {
         throw new NullPointerException("Name must not be null");
      }
      
      // Get the variable pool for this scope, creating one if necessary.
      Map pool = (Map) variables.get(scope);
      if (pool == null)
      {
         pool = new LinkedHashMap();
         variables.put(scope, pool);
      }
      
      // Disallow duplicate variable names in the same scope.
      if (pool.containsKey(name))
      {
         throw new SymbolException(
            "Duplicate user variable '" + name + "' declared in scope:  " +
            scope);
      }
      
      Variable var = null;
      try
      {
         var = new Variable(name, type, initializer, readOnly);
      }
      catch (IllegalArgumentException exc)
      {
         throw new SymbolException(
            "Error declaring user variable in scope:  " + scope, exc);
      }
      
      pool.put(name, var);
      
      return var;
   }
   
   /**
    * Retrieve a variable with the given name which is registered within the
    * context of the given scope.
    *
    * @param   scope
    *          Scope under which variable is registered.
    * @param   name
    *          Name under which variable is registered.
    *
    * @return  The specified variable, or <code>null</code> if none was
    *          found.
    */
   public final synchronized Variable getVariable(Scope scope, String name)
   {
      Map pool = (Map) variables.get(scope);
      if (pool != null)
      {
         return (Variable) pool.get(name);
      }
      
      return null;
   }
   
   /**
    * Drop a particular variable from a variable pool.  This permanently
    * removes the variable from the pool at the given scope.
    * 
    * @param   scope
    *          Scope with which target variable pool is associated.
    * @param   name
    *          Name of variable to be dropped.
    */
   public final synchronized void dropVariable(Scope scope, String name)
   {
      Map pool = (Map) variables.get(scope);
      if (pool != null)
      {
         pool.remove(name);
      }
   }
   
   /**
    * Drop a pool of variables.
    * 
    * @param   scope
    *          Scope with which the target variable pool is associated.
    */
   public final synchronized void dropVariablePool(Scope scope)
   {
      variables.remove(scope);
   }
   
   /**
    * Reset all variables in the specified scope to the values generated by
    * their respective initializer expressions, or to <code>null</code>. If a
    * variable has no associated initializer expression or the parameter
    * <code>nullify</code> is set to true, that variable's value is reset to
    * <code>null</code>.
    * <p>
    * The current scope of the symbol resolver is temporarily overridden to
    * be the specified scope, such that the initializer expressions
    * associated with the variables in this pool are executed in the proper
    * scope.
    * <p>
    * Currently, variables are not reset recursively within nested scopes;
    * only the variables directly associated with <code>scope</code> are
    * reset.
    *
    * @param   scope
    *          Scope for which variables are to be reset.
    * @param   nullify
    *          If <code>true</code>, null out variables' values;  else use
    *          the variables' default reset behavior.
    */
   public final synchronized void resetVariables(Scope scope, boolean nullify)
   {
      try
      {
         currentScope = scope;
         Map pool = (Map) variables.get(scope);
         if (pool != null)
         {
            Iterator iter = pool.values().iterator();
            while (iter.hasNext())
            {
               Variable var = (Variable) iter.next();
               if (nullify)
               {
                  var.setValue(null);
               }
               else
               {
                  var.reset();
               }
            }
         }
      }
      finally
      {
         currentScope = null;
      }
   }
   
   /**
    * Check whether a user expression has permission to invoke the specified
    * method on the specified target class. This method is called during
    * expression parsing when a symbol is encountered which would trigger a
    * method invocation at expression execution time.
    * <p>
    * This method is called only while an expression is parsed, never during
    * expression execution. It may be called multiple times for the same
    * combination of <code>target</code> and <code>method</code> while
    * parsing a single expression, if a symbol representing that method
    * invocation is encountered multiple times in that expression.
    * <p>
    * If permission is denied with a return of <code>false</code>, a
    * <code>SecurityException</code> will be thrown during expression
    * parsing, and the expression will not be compiled or executed.
    * <p>
    * <b>The default implementation of this method always returns
    * <code>true</code>; this method must be overridden to implement a
    * stricter security policy.</b>
    *
    * @param   target
    *          Class containing <code>method</code>.
    * @param   method
    *          Method which would be invoked at expression execution time
    *          if permission is granted.
    * @param   registered
    *          <code>true</code> if the method is registered as part of a
    *          function library; <code>false</code> if the method
    *          is one that is invoked explicitly on an object symbol within
    *          a user expression.
    *
    * @return  <code>true</code> to grant permission; <code>false</code> to
    *          deny it. This determination is completely at the discretion
    *          of the implementing class, based upon its required security
    *          policy.
    */
   protected boolean checkPermission(Class<?> target,
                                     Method method,
                                     boolean registered)
   {
      return true;
   }
   
   /**
    * This method should return the scope currently associated with the
    * symbol resolver. Scope is any object which implements the {@link
    * Scope} interface, but its implementation is completely up to the
    * concrete implementation of this class. This method is called at
    * the moment an expression is first executed, during parsing and
    * compilation.
    *
    * @return  An object which defines the current scope of the symbol
    *          resolver.
    */
   protected abstract Scope getCurrentScope();
   
   /**
    * Overriding implementations of this method should resolve the specified
    * string constant to a literal object. The return value must be of one
    * of the following types:
    * <ul>
    * <li>java.lang.String
    * <li>java.lang.Boolean
    * <li>java.lang.Integer
    * <li>java.lang.Long
    * <li>java.lang.Double
    * <li>java.lang.Float
    * </ul>
    * <p>
    * A return of any other type will cause a parsing error.
    * <p>
    * This method is invoked once per expression, at the time the expression
    * is first parsed by the {@link ExpressionParser}. It is invoked for a
    * string which is <b>not</b> enclosed in any quotation marks, which is
    * embedded within an expression.
    *
    * @param   qualifier
    *          An optional qualifier for concrete implementations which must
    *          support multiple namespaces. A qualifier is not mandatory for
    *          constants, so this parameter may be <code>null</code>.
    * @param   constant
    *          The string constant to be resolved to a literal.
    *
    * @return  The object represented by the constant <code>symbol</code>, or
    *          <code>null</code> if the symbol is not recognized.
    */
   protected Object resolveConstant(String qualifier, String constant)
   {
      return null;
   }
   
   /**
    * Retrieve a callback library object with the given name which is
    * registered within the context of the given scope.
    *
    * @param   scope
    *          Scope under which library is registered.
    * @param   name
    *          Name under which library is registered.
    *
    * @return  The specified library, or <code>null</code> if none was
    *          found.
    */
   final Object getLibraryTarget(Scope scope, String name)
   {
      Map library = (Map) libraries.get(scope);
      if (library != null)
      {
         return library.get(name);
      }
      
      return null;
   }
   
   /**
    * Retrieve a callback library object of the given type which is registered
    * within the context of the given scope.
    *
    * @param   scope
    *          Scope under which library is registered.
    * @param   cls
    *          Type of the library.
    *
    * @return  The specified library, or <code>null</code> if none was
    *          found.
    */
   final Object getLibraryTarget(Scope scope, Class<?> cls)
   {
      Map<Class<?>, Object> library = libsByClass.get(scope);
      if (library == null)
      {
         return null;
      }
      
      return library.get(cls);
   }
   
   /**
    * Resolve the symbol <code>name</code> to a Java bean-style "getter" or
    * "setter" method, where <code>name</code> represents the name of the
    * object being accessed or set. This may be (in the order of matching
    * precedence):
    * <ol>
    * <li>a callback library (in this case the setter semantic is
    *     disallowed, since a registered library may not be set to any other
    *     object by an expression);
    * <li>a property exposed by a callback library; or
    * <li>a user variable (setter semantic is allowed only if the variable
    *     was not registered as read-only).
    * </ol>
    * <p>
    * The search for a matching method commences in the resolver's {@link
    * #getCurrentScope current scope} and continues in {@link
    * Scope#getEnclosingScope enclosing scopes} as necessary, until a match
    * is found, or no more enclosing scopes exist.
    * <p>
    * This method is called by the {@link ExpressionParser} when a symbol
    * is encountered which could not be resolved as a constant.
    *
    * @param   function
    *          Function object which contains the class upon which the
    *          getter/setter will be invoked. May be <code>null</code>, in
    *          which case it is assumed the method is invoked on a registered
    *          callback library.
    * @param   name
    *          Name of the object for which the getter or setter method is
    *          required. Should begin with a lower case character.
    * @param   isSetter
    *          <code>true</code> if looking up the setter method;
    *          <code>false</code> if looking up the getter method.
    *
    * @return  A function object which holds the <code>Method</code> which
    *          will be invoked, the class of the target object upon which
    *          the method will be invoked, and the target object reference.
    *          If none can be found in the current scope, or in progressively
    *          broader enclosing scopes, <code>null</code> is returned.
    */
   final synchronized Function resolveFunction(Function function,
                                               String name,
                                               boolean isSetter)
   {
      Function result = null;
      
      Scope scope = (currentScope != null ? currentScope : getCurrentScope());
      while (scope != null && result == null)
      {
         result = resolveFunction(function, name, isSetter, scope);
         scope = scope.getEnclosingScope();
      }
      
      return result;
   }
   
   /**
    * Obtain a cached expression instance if the given key exists in the
    * cache at the given scope.
    * 
    * @param   scope
    *          Scope at which the target expression would be cached.
    * @param   key
    *          Used to lookup the instance in our cache.
    *
    * @return  The instance found or <code>null</code> if no instance is
    *          mapped to the given key.
    */
   CompiledExpression getCachedInstance(Scope scope, Object key)
   {
      try
      {
         instCacheLock.lock();
         
         Map<Object, CompiledExpression> map = instCache.get(scope);
         
         return (map == null ? null : map.get(key)); 
      }
      finally
      {
         instCacheLock.unlock();
      }
   }
   
   /**
    * Add a cached expression instance to the cache and map it with the
    * given key at the given scope.  If the key already exists in the cache,
    * the instance that is mapped will be replaced.
    *
    * @param   scope
    *          Scope at which the target expression is to be cached.
    * @param   key
    *          Used to lookup the instance in our cache.
    * @param   ce
    *          The instance to store.
    */
   void putCachedInstance(Scope scope, Object key, CompiledExpression ce)
   {
      try
      {
         instCacheLock.lock();
         
         Map<Object, CompiledExpression> map = instCache.get(scope);
         if (map == null)
         {
            map = new HashMap<Object, CompiledExpression>();
            instCache.put(scope, map);
         }
         
         map.put(key, ce);
      }
      finally
      {
         instCacheLock.unlock();
      }
   }
   
   /**
    * Lock the compiled expression instance cache.  Must be paired with a
    * subsequent call to {@link #unlockInstanceCache()}.
    */
   void lockInstanceCache()
   {
      instCacheLock.lock();
   }
   
   /**
    * Unock the compiled expression instance cache.  Must be paired with a
    * previous call to {@link #lockInstanceCache()}.
    */
   void unlockInstanceCache()
   {
      instCacheLock.unlock();
   }
   
   /**
    * Resolve the symbol <code>name</code> to a Java bean-style "getter" or
    * "setter" method, where <code>name</code> represents the name of the
    * object being accessed or set. This may be (in the order of matching
    * precedence):
    * <ol>
    * <li>a callback library (in this case the setter semantic is
    *     disallowed, since a registered library may not be set to any other
    *     object by an expression);
    * <li>a property exposed by a callback library; or
    * <li>a user variable (setter semantic is allowed only if the variable
    *     was not registered as read-only).
    * </ol>
    * <p>
    * The search for a matching method is limited to the specified scope.
    *
    * @param   function
    *          Function object which contains the class upon which the
    *          getter/setter will be invoked. May be <code>null</code>, in
    *          which case it is assumed the method is invoked on a registered
    *          callback library.
    * @param   name
    *          Name of the object for which the getter or setter method is
    *          required. Should begin with a lower case character.
    * @param   isSetter
    *          <code>true</code> if looking up the setter method;
    *          <code>false</code> if looking up the getter method.
    * @param   scope
    *          Scope to be searched.
    *
    * @return  A function object which holds the <code>Method</code> which
    *          will be invoked, the class of the target object upon which
    *          the method will be invoked, and the target object reference.
    *          If none can be found in the current scope, <code>null</code>
    *          is returned.
    */
   private Function resolveFunction(Function function,
                                    String name,
                                    boolean isSetter,
                                    Scope scope)
   {
      Function result = null;
      
      if (function == null)
      {
         // If name represents a library target, this is a special case
         // where the resolved function will be based on a static method
         // inside CompiledExpression, which accesses the target library
         // at runtime:  CompiledExpression.getLib(int).
         Object target = getLibraryTarget(scope, name);
         if (target != null)
         {
            // Library references are read-only;  can't set them.
            if (isSetter)
            {
               throw new UnsupportedOperationException(
                  "'" + name +
                  "' refers to a read-only library;  assignment not supported");
            }
            
            result = new Function(CompiledExpression.class,
                                  target.getClass(),
                                  METH_GET_LIB);
            result.setType(Function.GET_LIB);
            result.setSymbol(name);
            
            return result;
         }
      }
      
      // Prepare name for a getter/setter lookup.
      StringBuilder buf = new StringBuilder();
      buf.append(Character.toUpperCase(name.charAt(0)));
      if (name.length() > 1)
      {
         buf.append(name.substring(1));
      }
      
      if (isSetter)
      {
         buf.insert(0, "set");
         Class<?>[] sig = { UnknownType.class };
         result = resolveFunction(function, buf.toString(), sig, scope);
      }
      else
      {
         String suffix = buf.toString();
         buf.insert(0, "get");
         Class<?>[] sig = {};
         result = resolveFunction(function, buf.toString(), sig, scope);
         if (result == null)
         {
            buf.setLength(0);
            buf.append("is")
               .append(suffix);
            result = resolveFunction(function, buf.toString(), sig, scope);
         }
      }
      
      // Last chance is a user variable.
      if (result == null && function == null)
      {
         Variable var = getVariable(scope, name);
         if (var != null)
         {
            // Resolved function is based on CompiledExpression's built-in
            // getVar(int) or setVar(int, Object) method.
            if (isSetter)
            {
               if (var.isReadOnly())
               {
                  throw new UnsupportedOperationException(
                     "Cannot assign to read-only variable '" + name + "'");
               }
               
               result = new Function(CompiledExpression.class,
                                     Void.TYPE,
                                     METH_SET_VAR);
               result.setType(Function.SET_VAR);
            }
            else
            {
               result = new Function(CompiledExpression.class,
                                     var.getType(),
                                     METH_GET_VAR);
               result.setType(Function.GET_VAR);
            }
            
            result.setSymbol(name);
         }
      }
      
      return result;
   }
   
   /**
    * Resolve the symbol <code>name</code> to a Java method with the given
    * signature, where <code>function</code>, if not <code>null</code>,
    * defines the method which will be invoked to retrieve the target object
    * upon which the method represented by <code>name</code> will be invoked.
    * <p>
    * The search for a matching method commences in the resolver's {@link
    * #getCurrentScope current scope} and continues in {@link
    * Scope#getEnclosingScope enclosing scopes} as necessary, until a match
    * is found, or no more enclosing scopes exist.
    * <p>
    * This method is called by the {@link ExpressionParser} when the syntax
    * which represents a method invocation is encountered in an expression.
    *
    * @param   function
    *          Object which holds information about the method which is
    *          invoked to retrieve the target object. If <code>null</code>,
    *          it is assumed that <code>name</code> represents a method in
    *          a registered callback library.
    * @param   name
    *          Full name of the target method, with proper case.
    * @param   signature
    *          An array of the classes which comprise the target method's
    *          parameter list. This represents the list of argument types
    *          as parsed from a user expression. If a <code>null</code>
    *          parameter is passed to a method/function in the expression,
    *          and there is no corresponding, explicit cast, the associated
    *          slot in the array will have no corresponding class. Also, for
    *          methods which simulate variable argument lists by using an
    *          array of <code>Object</code>s as the last parameter, the
    *          lookup logic takes into account that this final argument
    *          has been "flattened out" in the user expression, such that
    *          we only see a one dimensional array of classes here.
    *
    * @return  A function object which holds the <code>Method</code> which
    *          will be invoked, the class of the target object upon which
    *          the method will be invoked, and the target object reference.
    *          If none can be found in the current scope, or in progressively
    *          broader enclosing scopes, <code>null</code> is returned.
    */
   final synchronized Function resolveFunction(Function function,
                                               String name,
                                               Class<?>[] signature)
   {
      Function result = null;
      
      Scope scope = (currentScope != null ? currentScope : getCurrentScope());
      while (scope != null && result == null)
      {
         result = resolveFunction(function, name, signature, scope);
         scope = scope.getEnclosingScope();
      }
      
      return result;
   }
   
   /**
    * Resolve the symbol <code>name</code> to a Java method with the given
    * signature, where <code>function</code>, if not <code>null</code>,
    * defines the method which will be invoked to retrieve the target object
    * upon which the method represented by <code>name</code> will be invoked.
    * <p>
    * The search for a matching method is limited to the specified scope.
    * <p>
    * This method is called by the {@link ExpressionParser} when the syntax
    * which represents a method invocation is encountered in an expression.
    *
    * @param   function
    *          Object which holds information about the method which is
    *          invoked to retrieve the target object. If <code>null</code>,
    *          it is assumed that <code>name</code> represents a method in
    *          a registered callback library.
    * @param   name
    *          Full name of the target method, with proper case.
    * @param   signature
    *          An array of the classes which comprise the target method's
    *          parameter list. This represents the list of argument types
    *          as parsed from a user expression. If a <code>null</code>
    *          parameter is passed to a method/function in the expression,
    *          and there is no corresponding, explicit cast, the associated
    *          slot in the array will have no corresponding class. Also, for
    *          methods which simulate variable argument lists by using an
    *          array of <code>Object</code>s as the last parameter, the
    *          lookup logic takes into account that this final argument
    *          has been "flattened out" in the user expression, such that
    *          we only see a one dimensional array of classes here.
    * @param   scope
    *          Scope to be searched.
    *
    * @return  A function object which holds the <code>Method</code> which
    *          will be invoked, the class of the target object upon which
    *          the method will be invoked, and the target object reference.
    *          If none can be found in the current scope, <code>null</code>
    *          is returned.
    */
   private Function resolveFunction(Function function,
                                    String name,
                                    Class<?>[] signature,
                                    Scope scope)
   {
      CacheKey key;
      
      // Handle the case of a direct method reference first.
      if (function != null)
      {
         Class<?> target = function.getReturnType();
         key = new CacheKey(scope, target, name, signature);
         
         return introspectFunction(key);
      }
      
      // Since that didn't work, it must be a library function, or nothing.
      Map map = (Map) libraryClasses.get(scope);
      if (map == null)
      {
         return null;
      }
      
      // Look in all registered libraries, but check for duplicates.
      Function func = null;
      Class<?> target = null;
      Iterator iter = map.keySet().iterator();
      while (iter.hasNext())
      {
         Class<?> next = (Class) iter.next();
         key = new CacheKey(scope, next, name, signature);
         Function tmp = introspectFunction(key);
         if (tmp != null)
         {
            if (func != null)
            {
               throw new AmbiguousSymbolException(
                  Function.asString(name, UnknownType.class, -1, signature));
            }
            func = tmp;
            target = next;
         }
      }
      
      if (func != null)
      {
         // Now we need to chain in the prerequisite function which serves
         // as the bootstrap for the user function we found. This gets
         // called on the compiled expression instance to retrieve the
         // library object on which the user function is called.
         Function prereq = new Function(CompiledExpression.class,
                                        target,
                                        METH_GET_LIB);
         prereq.setType(Function.GET_LIB);
         prereq.setSymbol((String) map.get(target));
         func.setPrerequisite(prereq);
      }
      
      return func;
   }
   
   /**
    * Use introspection to find a specific method within a specific target
    * class by an exact match of the method's name and an exact or fuzzy
    * match of the method's parameter signature. The first unambiguous match
    * is returned.
    * <p>
    * A cache is employed to short-circuit the search logic and quickly
    * provide a return if we have conducted a search with the same criteria
    * before. The cache records criteria for <em>failed</em> searches as
    * well as those for successful searches, so that we have a quick
    * response in either case.
    * <p>
    * If a method match is found, an {@link #accessCheck access check} is
    * conducted to ensure we have permission to compile this method
    * invocation into the compiled expression. By default, this check will
    * always pass, unless the concrete, symbol resolver subclass has
    * overridden {@link #checkPermission} to behave differently.
    * <p>
    * The search for a method match considers that the expression may be
    * using a variable length argument list. This is only the case when
    * the invocation target is a registered callback library. For targets
    * which are not registered callback library objects, a reflexive method
    * lookup which assumes an exact parameter signature match is attempted
    * first. If this fails, or if the target class is the type of a
    * registered callback library, several "fuzzy" signature matches are
    * attempted, in the following order:
    * <ol>
    * <li>A "strict" signature match: the method name must be an exact
    *     match; each required parameter type must be assignable from the
    *     provided parameter type, OR they must be the same primitive types
    *     or at least represent a compatible primitive/wrapper pair (e.g.,
    *     <code>int</code> and <code>java.lang.Integer</code>).
    * <li>A "lenient" signature match: the method name must be an exact
    *     match; each required parameter type must be assignable from the
    *     provided parameter type, OR they must be compatible boolean types
    *     (primitive/wrapper combination OK), OR they must be any combination
    *     of numeric primitive and/or wrapper types.
    * </ol>
    * <p>
    * If the invocation target type is an interface, the above sequence is
    * performed first for all methods of the interface (and all
    * superinterfaces). If this fails to produce a match, the sequence is
    * repeated for the <code>java.lang.Object</code> class.
    * <p>
    * Whether or not a method is detected, the result is stored in the lookup
    * cache under the provided <code>CacheKey</code> (<code>null</code> is
    * stored for a failed lookup).
    *
    * @param   key
    *          A cache key which defines scope, invocation target type,
    *          method name, and method signature.
    *
    * @return  A function object which holds the <code>Method</code> which
    *          will be invoked, the class of the target object upon which
    *          the method will be invoked, and class of the object which
    *          is expected to be returned from the method invocation. If no
    *          match is found, <code>null</code> is returned.
    *
    * @throws  SecurityException
    *          if the method invocation is disallowed by the security policy
    *          of the concrete implementation of this class.
    * @throws  AmbiguousSymbolException
    *          if any single application of a strict or lenient signature
    *          match algorithm against the list of methods in the target
    *          class being introspected detects more than one method which
    *          matches the search criteria.
    */
   private Function introspectFunction(CacheKey key)
   {
      // Check the cache first.
      Map cache = (Map) cacheRef.get();
      if (cache != null && cache.containsKey(key))
      {
         return (Function) cache.get(key);
      }
      
      Function func = null;
      
      Scope scope = key.scope;
      Class<?> target = key.target;
      String name = key.name;
      Class<?>[] signature = key.signature;
      
      boolean registered = isRegisteredLibrary(scope, target);
      
      // Try the simple, exact match first. We can only do this for targets
      // which are not registered libraries, because we need to support
      // var-args for those, which requires a more complicated match
      // algorithm.
      if (!registered)
      {
         try
         {
            Method method = target.getMethod(name, signature);
            func = new Function(target, method.getReturnType(), method);
         }
         catch (NoSuchMethodException exc)
         {
            // This just means the exact match did not work, so we need
            // to do a fuzzy signature match.
         }
      }
      
      // Scan each method to test for null matches and assignment
      // compatible parameters, as well as var-arg matches if dealing
      // with a registered library.
      if (func == null)
      {
         // Try strict signature match first.
         func = matchTargetMethod(target, name, signature, registered, true);
         
         // Try more lenient signature match if necessary.
         if (func == null)
         {
            func = matchTargetMethod(target,
                                     name,
                                     signature,
                                     registered,
                                     false);
         }
         
         // If target is an interface, check Object class for a possible
         // method match.
         if (func == null && target.isInterface())
         {
            // Strict match.
            func = matchTargetMethod(Object.class,
                                     name,
                                     signature,
                                     registered,
                                     true);
            
            // Lenient match.
            if (func == null)
            {
               func = matchTargetMethod(Object.class,
                                        name,
                                        signature,
                                        registered,
                                        false);
            }
         }
      }
      
      if (func != null)
      {
         // Before we use this function, first do a security check.
         accessCheck(func, registered);
      }
      
      // Cache the result (null or not) for faster lookup later, whether we
      // found a match or not.
      if (cache == null)
      {
         cache = new HashMap();
         cacheRef = new SoftReference(cache);
      }
      cache.put(key, func);
      
      return func;
   }
   
   /**
    * Introspect the specified target class to try to match a method with
    * the specified name and signature. Employs a "strict" or a "lenient"
    * match algorithm, depending upon the <code>strict</code> parameter:
    * <ul>
    * <li>A "strict" signature match: the method name must be an exact
    *     match; each required parameter type must be assignable from the
    *     provided parameter type, OR they must be the same primitive types
    *     or at least represent a compatible primitive/wrapper pair (e.g.,
    *     <code>int</code> and <code>java.lang.Integer</code>).
    * <li>A "lenient" signature match: the method name must be an exact
    *     match; each required parameter type must be assignable from the
    *     provided parameter type, OR they must be compatible boolean types
    *     (primitive/wrapper combination OK), OR they must be any combination
    *     of numeric primitive and/or wrapper types.
    * </ul>
    *
    * @param   target
    *          Method invocation target; the class being introspected for a
    *          matching method.
    * @param   name
    *          Exact (case-sensitive) name of the method to be matched.
    * @param   signature
    *          Array of types of the parameters being passed to the target
    *          method. May contain <code>null</code> elements if the type
    *          could not be determined.
    * @param   registered
    *          <code>true</code> if <code>target</code> is the class of a
    *          registered callback library; else <code>false</code>.
    * @param   strict
    *          <code>true</code> to perform a strict signature match;
    *          <code>false</code> to perform a lenient signature match.
    *
    * @return  A function object which holds the <code>Method</code> which
    *          will be invoked, the class of the target object upon which
    *          the method will be invoked, and class of the object which
    *          is expected to be returned from the method invocation. If no
    *          match is found, <code>null</code> is returned.
    *
    * @throws  AmbiguousSymbolException
    *          if more than one method which matches the search criteria is
    *          detected.
    */
   private Function matchTargetMethod(Class<?> target,
                                      String name,
                                      Class<?>[] signature,
                                      boolean registered,
                                      boolean strict)
   {
      Function func = null;
      
      Method[] methods = target.getMethods();
      Set candidates = new HashSet();
      int len = methods.length;
      for (int i = 0; i < len; i++)
      {
         Method next = methods[i];
         if (name.equals(next.getName()))
         {
            candidates.add(new Function(target,
                                        next.getReturnType(),
                                        next,
                                        signature,
                                        registered));
         }
      }
      
      Iterator iter = candidates.iterator();
      while (iter.hasNext())
      {
         Function next = (Function) iter.next();
         if (next.matchSignature(signature, strict))
         {
            if (func != null)
            {
               // Can't tell which one is the real match...
               throw new AmbiguousSymbolException(
                  target.getName() + "." + name);
            }
            func = next;
         }
      }
      
      return func;
   }
   
   /**
    * Performs an access check to obtain permission to compile a method
    * invocation into a compiled expression.
    *
    * @param   func
    *          Object from which the invocation target and method are
    *          retrieved.
    * @param   registered
    *          <code>true</code> if the invocation target represents a
    *          registered callback library; <code>false</code> if it
    *          represents an unregistered class.
    *
    * @throws  SecurityException
    *          if permission to compile the method invocation into the
    *          expression is denied.
    */
   private void accessCheck(Function func, boolean registered)
   {
      Method method = func.getMethod();
      if (!checkPermission(func.getTarget(), method, registered))
      {
         throw new SecurityException(
            "Invocation from an expression disallowed:  " + method);
      }
   }
   
   /**
    * Convenience method to determine whether <code>cls</code> represents
    * a callback library which has previously been registered in
    * <code>scope</code>.
    *
    * @param   scope
    *          Scope in which to check.
    * @param   cls
    *          The class for which we are searching.
    *
    * @return  <code>true</code> if <code>cls</code> was found in
    *          <code>scope</code>, else <code>false</code>.
    */
   private boolean isRegisteredLibrary(Scope scope, Class<?> cls)
   {
      Map map = (Map) libraryClasses.get(scope);
      
      return (map != null && map.containsKey(cls));
   }
   
   /**
    * A debug method which dumps the contents of the combined pool of
    * variables (across all scopes) to <code>stderr</code>.
    */
   void dumpVariablePool()
   {
      System.err.println("Variable pool:  " + variables);
   }
   
   /**
    * Helper class which is used to assist in lookups of methods when
    * introspecting invocation target classes. Instances of this class are
    * used as keys in the method lookup cache.
    */
   private static class CacheKey
   {
      /** Scope related to target class */
      private final Scope scope;
      
      /** Invocation target class being introspected */
      private final Class<?> target;
      
      /** Method name */
      private final String name;
      
      /** Method parameter signature */
      private final Class<?>[] signature;
      
      /** Cached hash code for {@link #signature} */
      private final int sigHash;
      
      /**
       * Convenience constructor which sets all fields relevant to this
       * key instance.
       *
       * @param   scope
       *          Scope.
       * @param   target
       *          Target class.
       * @param   name
       *          Method name.
       * @param   signature
       *          Method parameter signature.
       */
      CacheKey(Scope scope,
               Class<?> target,
               String name,
               Class<?>[] signature)
      {
         this.scope = scope;
         this.target = target;
         this.name = name;
         this.signature = signature;
         
         int len = signature.length;
         int hash = 17;
         for (int i = 0; i < len; i++)
         {
            hash = 37 * hash + (signature[i] == null
                                ? 0
                                : signature[i].hashCode());
         }
         sigHash = hash;
      }
      
      /**
       * Generate a well-distributed hash code value for this key instance.
       *
       * @return  Hash code.
       */
      public int hashCode()
      {
         int result = 17;
         result = 37 * result + scope.hashCode();
         result = 37 * result + target.hashCode();
         result = 37 * result + name.hashCode();
         result = 37 * result + sigHash;
         
         return result;
      }
      
      /**
       * Test for equivalence between this key and another.
       *
       * @param   o
       *          Other key to test.
       *
       * @return  <code>true</code> if this key instance is equivalent to
       *          <code>o</code>, else <code>false</code>.
       */
      public boolean equals(Object o)
      {
         // Assume cast is safe.
         CacheKey that = (CacheKey) o;
         
         return (scope.equals(that.scope)       &&
                 target.equals(that.target)     &&
                 name.equals(that.name)         &&
                 Arrays.equals(signature, that.signature));
      }
   }
   
   /**
    * Concrete implementation of a <code>SymbolResolver</code> for simple
    * testing and debug purposes.
    */
   private static class TestResolver
   extends SymbolResolver
   {
      private Scope testScope = null;
      
      private Object testLib = new Library();
      
      TestResolver()
      {
         super();
         
         testScope = new Scope()
         {
            public Scope getEnclosingScope()
            {
               return null;
            }
            public String toString()
            {
               return "test scope";
            }
         };
         
         registerLibrary(testScope, "test", testLib);
         
         Variable v;
         
         v = registerVariable(testScope, "var1", Integer.class, null, true);
         v.setValue(new Integer(42));
         
         v = registerVariable(testScope, "var2", Double.class, null, false);
         v.setValue(new Double(-745.34));
         
         v = registerVariable(testScope, "var3", String.class, null, false);
         v.setValue("Hello World!");
         
         v = registerVariable(testScope, "var4", Boolean.class, null, false);
         v.setValue(Boolean.TRUE);
         
         v = registerVariable(testScope, "var5", String.class, null, false);
         
         v = registerVariable(testScope, "var6", Integer.class, null, false);
         
         v = registerVariable(testScope, "i", Integer.class, null, false);
         v.setValue(new Integer(1));
         
         v = registerVariable(testScope, "$1", Integer.class, null, false);
         v.setValue(new Integer(5));
         
         v = registerVariable(testScope, "$2", String.class, null, false);
         v.setValue("second");
         
         v = registerVariable(testScope, "$3", Double.class, null, false);
         v.setValue(new Double(99.44));
      }
      
      public Object resolveConstant(String qualifier, String symbol)
      {
         if ("whatever".equals(symbol))
         {
            return "everwhat";
         }
         
         if ("constant".equals(symbol))
         {
            return symbol;
         }
         
         return super.resolveConstant(qualifier, symbol);
      }
      
      public boolean checkPermission(Class target,
                                     Method method,
                                     boolean registered)
      {
         if (testLib.getClass().equals(target) &&
             "illegal".equals(method.getName()))
         {
            return false;
         }
         
         return super.checkPermission(target, method, registered);
      }
      
      public Scope getCurrentScope()
      {
         return testScope;
      }
      
      public class Library
      {
         private boolean boolMember = false;
         
         public Boolean isBoolMember()
         {
            return (boolMember ? Boolean.TRUE : Boolean.FALSE);
         }
         public void setBoolMember(boolean boolMember)
         {
            this.boolMember = boolMember;
         }
         public Object getBoolAsObject()
         {
            return Boolean.TRUE;
         }
         public Integer getNum()
         {
            return new Integer(5);
         }
         public long getLongNum()
         {
            return (Integer.MAX_VALUE + 1L);
         }
         public Long getLongNumWrapped()
         {
            return new Long(getLongNum());
         }
         public Object getNumObject()
         {
            return new Integer(800);
         }
         public Object getLongNumObject()
         {
            return new Long(900L);
         }
         public int acceptInteger(Integer num)
         {
            return num.intValue();
         }
         public long acceptLong(Long num)
         {
            return num.longValue();
         }
         public String hello()
         {
            return "Hello";
         }
         public int varArgTest(Comparable comp,
                               char[] array,
                               int num,
                               Object[] args)
         {
            return (array.length + num + args.length);
         }
         public String primitiveArrayTest(Integer num, char[] array)
         {
            return (new String(array).toUpperCase());
         }
         public void illegal()
         {
         }
         public Object getNullValue()
         {
            return null;
         }
         public String string(long value)
         {
            return String.valueOf(value);
         }
         public String string(double value)
         {
            return String.valueOf(value);
         }
         public String string(boolean value)
         {
            return String.valueOf(value);
         }
         public int throwException()
         throws Exception
         {
            throw new Exception("Error!!!");
         }
      }
   }
}
