/*
** Module   : MatchCategory.java
** Abstract : counts or accumulates matches to a specific condition in some
**            list of files
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 GES 20050225   @20123 Created initial version supporting 
**                           maintaining a a map of files and for each
**                           file a set of matches.  Counters are kept 
**                           for the total number of matches, the
**                           number of files with matches and all of the 
**                           backing data is made accessible.
** 002 GES 20050318   @20440 Added support to save off the condition
**                           during match processing and to provide an
**                           iterator to these strings for future
**                           reporting use.
** 003 ECF 20111206          Added ID for database support.
** 004 GES 20130131          Cut down to work as little more than a container.
**                           Renamed the class to be a more accurate description.
*/ 

package com.goldencode.p2j.pattern;

import java.util.*;

/**
 * Counts the number of matches of a specific named condition by file. 
 */
public class MatchCategory
{
   /** Match category ID. */
   private int cid;
   
   /** The verbose descriptive name of this category, used in reports. */
   private String descr = null;
   
   /** Counter that records the total number of matches across all files. */
   private int aggregate = 0;

   /** Counter that records the total number files with matches. */
   private int matchFiles = 0;

   /** The list of file IDs for each of which an associated set of matches is stored. */
   private Map<Integer, Integer> fids = null;
   
   /**
    * Constructs an instance of a cross-file category with a specific
    * descriptive name.
    *
    * @param    cid
    *           MatchCategory ID.
    * @param    descr
    *           Verbose description of this category.
    * @param    aggregate
    *           The total number of matches to this category.
    * @param    matchFiles
    *           The total number of files with at least 1 match.
    * @param    fids
    *           The map of all file IDs and the number of matches for that file.
    */
   public MatchCategory(int                   cid,
                        String                descr,
                        int                   aggregate,
                        int                   matchFiles,
                        Map<Integer, Integer> fids)
   {
      this.cid        = cid;
      this.descr      = descr;
      this.aggregate  = aggregate;
      this.matchFiles = matchFiles;
      this.fids       = fids;
   }
   
   /**
    * Accesses the database ID for this category
    * 
    * @return   The database ID for this category.
    */
   public int getID()
   {
      return cid;
   }
   
   /**
    * Accesses the verbose description of this category.
    *
    * @return   The verbose description of this category.
    */
   public String getDescr()
   {
      return descr;
   }
   
   /**
    * Accesses the total number of matches across all files.
    *
    * @return   The total number of matches across all files.
    */
   public int getAggregateCounter()
   {
      return aggregate;
   }
   
   /**
    * Accesses the number of files which have matches.  Some files in
    * the list may have no matches.
    *
    * @return   The number of files in the list which have matches.
    */
   public int getMatchFiles()
   {
      return matchFiles;
   }
   
   /**
    * Accesses the iterator to the list of file-specific match sets.  This
    * allows direct access to the <code>Set</code> objects that each contain
    * all matches for each given file. 
    *
    * @return   The iterator to the list of match sets.
    */
   public Iterator<Integer> getFileIterator()
   {
      return fids.keySet().iterator();
   }
   
   /**
    * Returns the number of matches for a given file.
    *
    * @return   The number of matches for a given file or 0 if the file is
    *           not in the file list.
    */
   public int getCounterFor(int fid)
   {
      Integer matchCount = fids.get(fid);
      
      // safety first
      if (matchCount == null)
      {
         return 0;
      }
      
      return matchCount;
   }
}