/*
** Module   : SchemaStatsHelper.java
** Abstract : helper to copy and link to schema source files
**
** Copyright (c) 2007-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -------------------Description--------------------
** 001 GES 20070716   @34522 Helper to copy and link to schema source
**                           files.
** 002 GES 20130206          Moved parent class to the pattern package.
*/

package com.goldencode.p2j.schema;

import java.util.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.uast.*;

/**
 * Helper to copy and link to Progress 4GL schema source files.
 */
public class SchemaStatsHelper
implements SourceFileHelper
{
   /** Column headings. */
   private static final String[] headings = new String[]
   {
      "Original Source File",
      "Schema AST"
   };
   
   /** Store a reverse lookup from the schema name to the source file. */
   private HashMap reverse = new HashMap();
   
   /**
    * Obtain the list of source file descriptions which is suitable for use
    * as the column headings in the source file index table.
    *
    * @return   Column headings.
    */
   public String[] getHeadings()
   {
      return headings;
   }
   
   /**
    * Provide the list of copy specifications for the given source file and
    * output directory.
    *
    * @param    schema
    *           Schema AST filename.
    * @param    outDir
    *           Root of the output directory where all copies must be made.
    *
    * @return   The list of source and target copy operations to be made.
    */
   public CopySpec[] getSpecifications(String schema, String outDir)
   {
      // safety first
      if (schema == null)
         return null;
      
      CopySpec[] specs = new CopySpec[2];
      
      SourceSpec ss = (SourceSpec) reverse.get(schema);
      
      String src = null;
      String tar = null;
      String scm = null;
      
      try
      {
         // create the master source filename
         src = Configuration.toFile(ss.source).getAbsolutePath();
         scm = Configuration.toFile(schema).getAbsolutePath();
      }
      catch (ConfigurationException ce)
      {
         return null;
      }
      
      if (!ss.df)
      {
         src = String.format("%s%s", src, AstGenerator.CACHE_POSTFIX);
         tar = String.format("%s%s%s.html",
                             outDir,
                             ss.source,
                             AstGenerator.CACHE_POSTFIX);
      }
      else
      {
         tar = String.format("%s%s.html",
                             outDir,
                             ss.source);
      }
      
      specs[0] = new CopySpec(src, tar, CopySpec.Type.TXT_TO_HTML);
      specs[1] = new CopySpec(scm, String.format("%s%s", outDir, schema));
      
      return specs;
   }
   
   /**
    * Create a URL to link to the given source file and line number.
    *
    * @param    relPath
    *           Relative path to the root of the output directory where the
    *           source files copies were made.
    * @param    schema
    *           Schema AST filename.
    * @param    line
    *           The line number to which to link.
    *
    * @return   A URL to link to the copied file.
    */
   public String getURL(String relPath, String schema, String line)
   {
      StringBuilder sb = new StringBuilder();
      SourceSpec spec = (SourceSpec) reverse.get(schema);
      
      sb.append(relPath)
        .append(spec.source);
        
      if (!spec.df)
         sb.append(AstGenerator.CACHE_POSTFIX);
      
      sb.append(".html#")
        .append(line);
      
      return sb.toString();
   }
   
   /**
    * Add a mapping definition from the given schema name to the original
    * source file that defined that schema. This must be obtained from the
    * schema AST files during any report processing.  This is relied upon
    * in order to properly determine the files for copying.
    *
    * @param    schema
    *           The schema AST file name.
    * @param    source
    *           The source file.
    * @param    df
    *           <code>true</code> if this source file is a .df file.
    */
   public void addMapping(String schema, String source, boolean df)
   {
      reverse.put(schema, new SourceSpec(source, df));
   }
   
   /**
    * Simple data structure to store key data for reverse mapping purposes.
    */
   private static class SourceSpec
   {
      /** Source file name. */
      private String source = null;
      
      /** <code>true</code> if the source file is a .df file. */
      private boolean df = false;
      
      /**
       * Create an instance.
       *
       * @param    source
       *           The source file.
       * @param    df
       *           <code>true</code> if this source file is a .df file.
       */
      SourceSpec(String source, boolean df)
      {
         this.source = source;
         this.df     = df;
      }
   }
}
